#include "display.h"
#include "views.h"
#include "canvas.h"
#include "recorder.h"
#include "winex.h"
#include "wassert.h"

Painter::Painter (
        HWND hwnd,
        ViewWave& viewWave,
        ViewFreq& viewFreq,
        int samplesPerBuf, 
        int samplesPerSec )
:   _hwnd (hwnd), 
    _viewWave (viewWave),
    _viewFreq (viewFreq),
    _samplesPerBuf (samplesPerBuf),
    _samplesPerSecond (samplesPerSec),
    _bitsPerSample (16),
    _pRecorder (samplesPerBuf, samplesPerSec)
{
    // Let the caller Resume explicitly
    //_thread.Resume ();
}

void Painter::Run ()
{
    for (;;)
    {
        _event.Wait ();
        if (_isDying)
            return;

        Lock lock (_mutex);
        if (_pRecorder->IsBufferDone ())
            LokWaveInData ();
    }
}

void Painter::FlushThread ()
{
    _event.Release ();
}

BOOL Painter::ReInit (
        int samplesPerBuf, 
        int samplesPerSec, 
        int bitsPerSample)
{
    Lock lock (_mutex);

    if (_pRecorder->BitsPerSample() == bitsPerSample &&
        _pRecorder->SamplesPerSecond() == samplesPerSec &&
        _pRecorder->SampleCount() == samplesPerBuf)
    {
        return TRUE;
    }

    _samplesPerBuf = samplesPerBuf;
    _samplesPerSecond = samplesPerSec;
    _bitsPerSample = bitsPerSample;

    BOOL isStarted = _pRecorder->IsStarted();
    if (isStarted)
        _pRecorder->Stop();

    BOOL isError = FALSE;

    try
    {
        _pRecorder.ReInit (_bitsPerSample, _samplesPerBuf, _samplesPerSecond);
    }
    catch ( WinException e )
    {
        char buf[50];
        wsprintf ( buf, "%s, Error %d", e.GetMessage(), e.GetError() );
        MessageBox (0, buf, "Exception", MB_ICONEXCLAMATION | MB_OK);
        isError = TRUE;
    }
    catch (...)
    {
        MessageBox (0, "Unknown", "Exception", MB_ICONEXCLAMATION | MB_OK);
        isError = TRUE;
    }

    if (isStarted)
    {
        isError = !_pRecorder->Start(_event);
    }

    return !isError;
}

BOOL Painter::Start ()
{
    Lock lock (_mutex);
    return _pRecorder->Start (_event);
}

void Painter::Stop ()
{
    Lock lock (_mutex);
    _pRecorder->Stop ();
}

void Painter::LokWaveInData ()
{
    SampleIter iter (_pRecorder.GetAccess());
    // Quickly release the buffer
    if (!_pRecorder->BufferDone ())
        return;

    _viewFreq.Update (iter);
    _viewWave.Update ();
}


